using System;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Channels.Tcp;
using System.Runtime.Remoting.Channels.Http;
using System.Reflection;
using System.Net.Sockets;
using System.Threading;

namespace SingletonLibrary
{
    public class Singleton : MarshalByRefObject
    {

        private static Singleton instance;

        private static HttpChannel channel;

        private static readonly Object threadlock = new Object();

        private Singleton()
        {
            //Constructor logic here
        }

        //To access HashCode of Singleton rather than his proxy
        public string GetRealHashCode()
        {
            return this.GetHashCode().ToString();
        }

        public bool CheckConnection()
        {
            //Don't really do anything
            return true; //since object is remote if this won't fail - we have connection
        }

        public static Singleton GetInstance()
        {
            lock (threadlock) 
            //as there might be multiple threads in same worker thread to instansiate singleton 
            {
                if (instance == null)
                {
                    try
                    {
                        //Attempting to create the Server - there might be a better way to check 
			  //if there is one already
                        channel = new HttpChannel(8089);
                    }
                    //Server already created just fetch the proxy to our singleton
                    catch (SocketException)
                    {
                        channel = new HttpChannel();
                        ChannelServices.RegisterChannel(channel, false);
                        //loop for few seconds - as it might be contructing.
                        for (int i = 0; i <= 6; i++)
                        {
                            try
                            {
                                instance = (Singleton)Activator.GetObject(typeof(Singleton), 							"http://localhost:8089/Singleton"); 
                                break;
                            }
                            catch (RemotingException)
                            {
                                //Server Channel Created but not rtg - this might happen when there
                                //is slow object construction
                                //We migh want to put delay here and attempt again
                                Thread.Sleep(300);
                            }
                        }
                        if (instance == null) throw new Exception(
							"Remote object isn't reachable - timeout");
                    }
                    if (instance == null) //Server needs to be initialised we cold move that piece 
					     //of code in the same channel = new TcpChannel(8080) 
                    {

                        ChannelServices.RegisterChannel(channel, false);
                        RemotingConfiguration.RegisterWellKnownServiceType(typeof(Singleton), 
							"Singleton", WellKnownObjectMode.Singleton);
                        //Now we have a remoted object so we need to get it
                        instance = (Singleton)Activator.GetObject(typeof(Singleton), 
								"http://localhost:8089/Singleton"); 
			  // We could provably use Instance to avoid duplication of code, but we would 
			  // run "new TcpChannel(8080)" again because of performance might be good to 
			  // repeat code now
                    }
                }

                try
                {
                    instance.CheckConnection(); 
		     //this proxy will throw socetexception if there is no connectivity
                }
                catch (SocketException)
                // That Means connection is no longer available - server must have died
                {
                    //We'll need a new one as we will become server now
                    ChannelServices.UnregisterChannel(channel); 
                    instance = null; 
                    return GetInstance(); 
                }
                return instance;
            }
        }
    }
}